#!/usr/bin/env python3
"""
Network Topology Mapper
Parameters:
    - IP: Starting IP address
    - subnet: Network subnet (CIDR notation)
Additional parameters in comments:
    - max_hops: Maximum number of hops to trace
    - output_format: Format for topology map (dot, json)
"""
import networkx as nx
import matplotlib.pyplot as plt
import scapy.all as scapy
from scapy.layers.inet import IP, ICMP, TCP
import argparse
import json
from pathlib import Path
import logging
import threading
import queue
import time
from datetime import datetime
import ipaddress

class NetworkTopologyMapper:
    def __init__(self, start_ip, subnet, max_hops=20, output_dir="topology_maps"):
        self.start_ip = start_ip
        self.subnet = subnet
        self.max_hops = max_hops
        self.output_dir = Path(output_dir)
        self.output_dir.mkdir(exist_ok=True)
        
        self.network = nx.Graph()
        self.discovered_devices = set()
        self.device_queue = queue.Queue()
        self.results_queue = queue.Queue()
        self.stop_event = threading.Event()
        
        self.setup_logging()
        
    def setup_logging(self):
        """Set up logging configuration"""
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[
                logging.FileHandler(self.output_dir / "mapper.log"),
                logging.StreamHandler()
            ]
        )
        
    def discover_devices(self):
        """Discover active devices in the network"""
        network = ipaddress.ip_network(self.subnet)
        logging.info(f"Scanning network {network}")
        
        # Create and send ARP requests
        arp = scapy.ARP(pdst=str(network))
        ether = scapy.Ether(dst="ff:ff:ff:ff:ff:ff")
        packet = ether/arp
        
        result = scapy.srp(packet, timeout=3, verbose=0)[0]
        
        devices = []
        for sent, received in result:
            devices.append({
                'ip': received.psrc,
                'mac': received.hwsrc
            })
            self.device_queue.put(received.psrc)
            self.discovered_devices.add(received.psrc)
            
        logging.info(f"Discovered {len(devices)} devices")
        return devices
        
    def traceroute(self, target_ip):
        """Perform traceroute to target"""
        hops = []
        for ttl in range(1, self.max_hops + 1):
            pkt = IP(dst=target_ip, ttl=ttl) / ICMP()
            
            # Send packet and wait for response
            reply = scapy.sr1(pkt, timeout=2, verbose=0)
            
            if reply is None:
                # No response
                hops.append(None)
            elif reply.type == 0:
                # Echo reply - reached target
                hops.append(reply.src)
                break
            else:
                # Time exceeded - intermediate hop
                hops.append(reply.src)
                
        return [hop for hop in hops if hop is not None]
        
    def scan_ports(self, ip, ports=[21, 22, 23, 80, 443]):
        """Scan common ports on target"""
        open_ports = []
        for port in ports:
            tcp_connect = IP(dst=ip)/TCP(dport=port, flags="S")
            response = scapy.sr1(tcp_connect, timeout=1, verbose=0)
            
            if response and response.haslayer(TCP):
                if response[TCP].flags == 0x12:  # SYN-ACK
                    open_ports.append(port)
                    
        return open_ports
        
    def get_device_info(self, ip):
        """Get detailed information about a device"""
        info = {'ip': ip, 'open_ports': []}
        
        try:
            # Try to get hostname
            try:
                hostname = socket.gethostbyaddr(ip)[0]
                info['hostname'] = hostname
            except:
                pass
                
            # Scan common ports
            info['open_ports'] = self.scan_ports(ip)
            
            # Try SNMP system description (if available)
            # This would require additional SNMP library
            
        except Exception as e:
            logging.error(f"Error getting info for {ip}: {str(e)}")
            
        return info
        
    def map_network(self):
        """Map the network topology"""
        logging.info("Starting network topology mapping")
        start_time = time.time()
        
        # Discover active devices
        devices = self.discover_devices()
        
        # Add devices to graph
        for device in devices:
            self.network.add_node(device['ip'], **self.get_device_info(device['ip']))
            
        # Map connections between devices
        mapped_pairs = set()
        
        while not self.device_queue.empty():
            source_ip = self.device_queue.get()
            
            # Perform traceroute to other discovered devices
            for target_ip in self.discovered_devices:
                if source_ip != target_ip and (source_ip, target_ip) not in mapped_pairs:
                    hops = self.traceroute(target_ip)
                    
                    # Add hops to graph
                    for i in range(len(hops) - 1):
                        if hops[i] and hops[i+1]:
                            self.network.add_edge(hops[i], hops[i+1])
                            
                    mapped_pairs.add((source_ip, target_ip))
                    mapped_pairs.add((target_ip, source_ip))
                    
        duration = time.time() - start_time
        logging.info(f"Mapping completed in {duration:.1f} seconds")
        
        # Save results
        self.save_results(duration)
        
    def create_graphviz(self):
        """Create Graphviz visualization"""
        try:
            plt.figure(figsize=(12, 8))
            pos = nx.spring_layout(self.network)
            
            # Draw nodes
            nx.draw_networkx_nodes(self.network, pos, node_color='lightblue',
                                node_size=500)
            
            # Draw edges
            nx.draw_networkx_edges(self.network, pos)
            
            # Add labels
            labels = {}
            for node in self.network.nodes():
                label = node
                if 'hostname' in self.network.nodes[node]:
                    label += f"\n{self.network.nodes[node]['hostname']}"
                if 'open_ports' in self.network.nodes[node]:
                    ports = self.network.nodes[node]['open_ports']
                    if ports:
                        label += f"\nPorts: {', '.join(map(str, ports))}"
                labels[node] = label
                
            nx.draw_networkx_labels(self.network, pos, labels)
            
            # Save plot
            plt.title("Network Topology Map")
            plt.axis('off')
            plt.savefig(self.output_dir / f"topology_{datetime.now().strftime('%Y%m%d_%H%M%S')}.png")
            plt.close()
            
        except Exception as e:
            logging.error(f"Error creating visualization: {str(e)}")
            
    def save_results(self, duration):
        """Save mapping results"""
        # Save network graph as JSON
        graph_data = {
            'nodes': [{'id': n, **self.network.nodes[n]} for n in self.network.nodes()],
            'edges': list(self.network.edges())
        }
        
        results = {
            'timestamp': datetime.now().isoformat(),
            'duration': duration,
            'start_ip': self.start_ip,
            'subnet': self.subnet,
            'devices_found': len(self.discovered_devices),
            'topology': graph_data
        }
        
        # Save JSON results
        json_file = self.output_dir / f"topology_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"
        with open(json_file, 'w') as f:
            json.dump(results, f, indent=2)
            
        logging.info(f"Results saved to {json_file}")
        
        # Create visualization
        self.create_graphviz()
        
    def generate_report(self):
        """Generate human-readable report"""
        report = []
        report.append("Network Topology Report")
        report.append("=" * 50)
        report.append(f"Scan started: {datetime.now().isoformat()}")
        report.append(f"Network: {self.subnet}")
        report.append(f"Starting IP: {self.start_ip}")
        report.append(f"Devices discovered: {len(self.discovered_devices)}")
        report.append("\nDevice Details:")
        report.append("-" * 50)
        
        for node in sorted(self.network.nodes()):
            report.append(f"\nDevice: {node}")
            node_data = self.network.nodes[node]
            if 'hostname' in node_data:
                report.append(f"Hostname: {node_data['hostname']}")
            if 'open_ports' in node_data:
                report.append(f"Open ports: {', '.join(map(str, node_data['open_ports']))}")
            neighbors = list(self.network.neighbors(node))
            if neighbors:
                report.append(f"Connected to: {', '.join(neighbors)}")
                
        # Save report
        report_file = self.output_dir / f"report_{datetime.now().strftime('%Y%m%d_%H%M%S')}.txt"
        with open(report_file, 'w') as f:
            f.write('\n'.join(report))
            
        logging.info(f"Report saved to {report_file}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Network Topology Mapper")
    parser.add_argument("ip", help="Starting IP address")
    parser.add_argument("subnet", help="Network subnet (CIDR notation)")
    parser.add_argument("--max-hops", type=int, default=20,
                      help="Maximum number of hops to trace")
    parser.add_argument("--output-dir", default="topology_maps",
                      help="Directory to store output files")
    args = parser.parse_args()
    
    mapper = NetworkTopologyMapper(args.ip, args.subnet, args.max_hops, args.output_dir)
    mapper.map_network()
    mapper.generate_report()
